/*******************************************************
 *
 *  File: a2kboot.c
 *
 *  Author: Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *  (c) Copyright 2001, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************/

#include <stdarg.h>
#include "a2kboot.h"
#include "video.h"
#include "lib.h"
#include "uart.h"
#include "cpu.h"
#include "version.h"
#include "crc.h"

#define KERNEL_PHYS_ADDRESS			0xf0008000
#define FRAMEBUFFER_PHYS_ADDRESS	0xfd000000

#define INITRD_PHYS_ADDRESS			0xf8100000

unsigned short* video = (unsigned short*) FRAMEBUFFER_PHYS_ADDRESS;


char* skip_white(char* buf)
{
	char	c;
	int		is_white;
	
	do
	{
		c = *buf;
		is_white = ((c == ' ') || (c=='\t') || (c=='\n'));
		if (is_white)
			buf++;
	} while (is_white);
	
	return buf;
}


void foo()
{
	int	i=0;
	for (i=0 ; i<640 ; i++)
		video[i]=0xf;
}


void uart_line_control(int uart, char* buf)
{
	unsigned reg = REG_UART_UARTFCR(uart);
	int		num;
	
	num = lib_bits(reg, 5, 6);
	lib_sprintf(buf, "  bits =%d, fifo=%s, 2stop bits=%s, even parity=%s, parity=%s, sndbreak=%s\n", 5 + num, lib_bit(reg, 4)?"yes" : "no",
		lib_bit(reg, 3)?"yes":"no", lib_bit(reg, 2)?"yes":"no", lib_bit(reg, 1)?"yes" : "no", lib_bit(reg, 0)?"yes":"no");
	
}

void uart_baudrate(int uart, char* buf)
{
	unsigned reg = REG_UART_UARTBRCR(uart);
	unsigned brdiv = lib_bits(reg, 0, 15);
	
	lib_sprintf(buf, "  baudrate div=%d", brdiv);
	switch (brdiv)
	{
		case 0:
			lib_strcat(buf, " (460800)\n");
			break;
		case 1:
			lib_strcat(buf, " (230400)\n");
			break;
		case 2:
			lib_strcat(buf, " (153600)\n");
			break;
		case 3:
			lib_strcat(buf, " (115200)\n");
			break;
		default:
			lib_strcat(buf, "\n");
			break;
	}
	
}

void uart_dump_info(int uart)
{
	char	buf[128];

	lib_sprintf(buf, "Uart %d\n", uart);
	video_puts(buf);

	uart_line_control(uart, buf);
	video_puts(buf);

	uart_baudrate(uart, buf);
	video_puts(buf);
}

void uart_printch(int uart, unsigned char c)
{
	while (((REG_UART_UARTFLG(uart)) & UARTFLG_TXFF ) == UARTFLG_TXFF)
		;
	REG_UART_UARTDR(uart) = c; 
}

unsigned char uart_getchar(int uart)
{
	while (((REG_UART_UARTFLG(uart)) & UARTFLG_RXFE ) == UARTFLG_RXFE)
		;
	return (REG_UART_UARTDR(uart)) & UARTDR_DATA; 
}


void uart_echo(int uart)
{
	unsigned char	c;
	int	done = 0;

	while(!done)
	{
		c = uart_getchar(uart);
		video_printch(c);
		
		done = (c == '=');
	}
}

void uart_drain_rx(int uart)
{
	volatile	char c;

	while (((REG_UART_UARTFLG(uart)) & UARTFLG_RXFE ) != UARTFLG_RXFE)
		c = (REG_UART_UARTDR(uart)) & UARTDR_DATA; 
}

void uart_puts(int uart, const char* s)
{
	while(*s)
		uart_printch(uart, *s++);
}


void uart_receive_textline(int uart, char* buf)
{
	unsigned char	c;
	
	do
	{
		c = uart_getchar(uart);
		*buf++ = c;
	} while (c != '\n');
	*(buf-1) = 0;
}


void uart_receive_data(int uart, unsigned char *dst, int size, int progress)
{
	int 			count = 0;
	unsigned char 	c;
#ifdef UART_TEST
	char*			testdata = "1234567890abcdef\n";
	char*			pc = testdata;
#endif
	
	while(size--)
	{
		c = uart_getchar(uart);
		count++;
		if (dst)
			*dst++ = c;
#ifdef UART_TEST
		if (c != *pc++)
		{
			video_puts("Error is data!\n");
			return;
		}
		if (!(*pc))
			pc = testdata;
#endif
		
		
		if (count == progress)
		{
			video_printch('.');
			count = 0;
		}
	}
}

void parse_file_info(char* buf, int* filesize, char** filename, unsigned* crc)
{
	buf = skip_white(buf);
	buf = lib_my_atoi(buf, filesize);
	buf = skip_white(buf);
	*filename = buf;
	while(*buf)
	{
		if (*buf++ == ' ')
		{
			*(buf-1) = 0;
			lib_my_atoi(buf, crc);
		}
	}
}

static void verify_crc(const char* filename, const char* dst, int filesize, unsigned crc)
{
	char	line[64];
	unsigned	crc_calc = 0;
	crc_calc = crc_data(dst, filesize, crc_calc);
	if (crc_calc != crc)
	{
		lib_sprintf(line, "Warning, %s CRC checks do not match, calculated: %i\n", filename, crc_calc);
		video_puts(line);
		uart_puts(3, line);
	}
	else
		video_puts("crc check ok\n");
}

int uart_receive_file(int uart, char* dst)
{
	char	buf[256];
	char	buf1[128];
	int		filesize = 0;
	char*	filename;
	unsigned	crc = 0;

	uart_receive_textline(uart, buf);
	parse_file_info(buf, &filesize, &filename, &crc);
	if (crc)
		lib_sprintf(buf1, "File=%s, size=%d, crc=%d\n", filename, filesize, crc);
	else
		lib_sprintf(buf1, "File=%s, size=%d (no crc check)\n", filename, filesize);
	video_puts(buf1);
	uart_receive_data(uart, dst, filesize, 10240);

	if (crc)
		verify_crc(filename, dst, filesize, crc);
	return filesize;
}


void check_initrd()
{
	char		line[128];
	char*		initrd_adr = (char*) INITRD_PHYS_ADDRESS+10;
	char		c;
	int			always_ask = 1;

		/*
		 * Ramdisks have magic name "ramdisk" at offset 10
		 */
	//if (lib_strncmp(initrd_adr, "ramdisk", 7) == 0)
	if (initrd_adr[0] == 'r' && initrd_adr[1] == 'a' && initrd_adr[2] == 'm'
		&& initrd_adr[3] == 'd' && initrd_adr[4] == 'i' && initrd_adr[5] == 's'
		&& initrd_adr[6] == 'k' && !always_ask)
	{
		uart_puts(3, "\n\rRamdisk still present\n\r");
		return;
	}
	uart_puts(3, "\n\rRamdisk not found, press 'y' to load one.\n\r");

	c = uart_getchar(3);
	if (c == 'y')
	{
		lib_sprintf(line, "Loading initrd image at %x...\n", INITRD_PHYS_ADDRESS);
		video_puts(line);
		uart_receive_file(3, (char*) INITRD_PHYS_ADDRESS);
	}
}


/*
 * We've restarted ourselves and we now run at 0xfc00.0000
 * caches are disabled, MMU is still on.
 */
void a2kboot_restart()
{
	char		line[128];
	unsigned 	phys_dst = KERNEL_PHYS_ADDRESS;

	video_setup((int*)FRAMEBUFFER_PHYS_ADDRESS, 640, 480, 12, 0xf, 0);
	video_clear();
	lib_sprintf(line, "Restarted...\n");
	//video_puts(line);
	//lib_sprintf(line, "CP15 control=%x, Local variable i=%x\n", _cp15_read_control_register(), &i);
	//video_puts(line);

	_disable_mmu();
	video_puts("MMU is off.\n");
	video_puts("A2kboot is now booting the Linux kernel...\n");

		/*
		 * this will speed up the Linux kernel uncompressing itself 
		 */
	_enable_I_cache();

	uart_puts(3, "\n\rA2kboot is now booting the Linux kernel...\n\n\r");
	video_puts("Bye...\n");
	_boot_linux_kernel(phys_dst);
	video_puts("Hmm still here...\n");
	while(1)
		;
}
void test()
{
	char		line[128];
	int			i=0;
	while(1)
	{
		lib_sprintf(line, "This is test line %d\n", i++);
		video_puts(line);
	}
}

/*
 * Main entry point for a2kboot. We run at 0x0180.0000, caches
 * are disabled and MMU is on.
 * To get full download speed we will enable caches.
 */
void a2kboot_main()
{
	char		line[128];
	unsigned 	phys_dst = KERNEL_PHYS_ADDRESS;

	cpu_init_stuff();

	_enable_caches();
	
	video_setup((int*)FRAMEBUFFER_PHYS_ADDRESS, 640, 480, 12, 0xf, 0);
	video_clear();
	video_puts("A2kboot: Agilent AAED2000 Linux boot loader\n");
	lib_sprintf(line, "         v%d.%d by Peter van Sebille (c) 2001\n\n", VERSION_MAJOR, VERSION_MINOR);
	video_puts(line);

	//lib_sprintf(line, "CP15 control=%x, Local variable i=%x\n", _cp15_read_control_register(), &i);
	//video_puts(line);

	//uart_dump_info(1);
	//uart_dump_info(2);
	//uart_dump_info(3);

	uart_drain_rx(3);

	lib_sprintf(line, "Loading kernel image at %x...\n", phys_dst);
	video_puts(line);
	uart_receive_file(3, (char*) phys_dst);
	video_puts("\n\n");

	check_initrd();

	_disable_caches();

	_a2kboot_restart();
}



